# Pytest suite for MP2. This program is provided to students to test their
# functions.
# Version: 24sp

# Note to students: This line is a convention in pytest suites, an
# exception to the poor practice of import * since we test the
# mp2_mastermind.py functions without qualifying them.
from mp2_mastermind import *


def test_make_random_code():
    """
    Tests the random_code() function for MP2, which should take no arguments
    and return a 4-character string randomly composed of the characters
    'RGBPOW'.
    """
    colors = 'RGBPOW'
    used = list(colors)
    for _ in range(200):
        code = make_random_code()
        assert isinstance(code, str)
        assert len(code) == 4
        for char in code:
            assert char in colors
            if char in used:
                # remove the character from used; we expect this to be empty
                # after 200 runs
                used.remove(char)

    assert len(used) == 0 # all characters should have been used at some point


def make_random_codes(n):
    """
    Helper function for generating n random mastermind codes to help with
    testing.

    Arguments:
        n (int): Number of codes to randomly-generate (should be > 0)

    Returns:
        (list): List of n random mastermind code strings.
    """
    codes = []
    for _ in range(n):
        # We assume that make_random_code() works.
        code = make_random_code()
        codes.append(code)
    return codes


def test_count_exact_matches():
    """
    Tests the student's count_exact_matches function. This function
    should take two 4-letter string arguments and return an int
    based on how many letters are shared in exact positions.
    """
    # Check basic invariants.
    codes1 = make_random_codes(1000)
    codes2 = make_random_codes(1000)
    for (c1, c2) in zip(codes1, codes2):
        # count_exact_matches should always be a value between
        # 0 (the '----' case) and 4 (the 'bbbb' case)
        assert 0 <= count_exact_matches(c1, c2) <= 4

    # Check specific answers.
    assert count_exact_matches('RGBP', 'RGBP') == 4
    assert count_exact_matches('RRRR', 'RRRR') == 4
    assert count_exact_matches('RRRR', 'OOOO') == 0
    assert count_exact_matches('RROO', 'OORR') == 0
    assert count_exact_matches('RORR', 'RROR') == 2
    assert count_exact_matches('RORP', 'RPOR') == 1
    assert count_exact_matches('RGRP', 'RGRR') == 3
    assert count_exact_matches('RGBP', 'GBPR') == 0
    assert count_exact_matches('RGBP', 'RGOO') == 2
    assert count_exact_matches('RGBP', 'RWBW') == 2
    assert count_exact_matches('RGBP', 'WOWO') == 0


def test_count_letter_matches():
    """
    Tests the student's count_letter_matches function. This function
    should take two 4-letter string arguments and return an int
    based on how many letters are shared.
    """
    # Check basic invariants.
    codes1 = make_random_codes(1000)
    codes2 = make_random_codes(1000)
    for (c1, c2) in zip(codes1, codes2):
        assert 0 <= count_letter_matches(c1, c2) <= 4

    # Check specific answers.
    # Sanity check that count_exact matches works.
    assert count_exact_matches('RGBP', 'RGBP') == 4
    # Next, test count_letter_matches.
    assert count_letter_matches('RGBP', 'RGBP') == 4
    assert count_letter_matches('RGBP', 'GBPR') == 4
    assert count_letter_matches('ROGO', 'RRGG') == 2
    assert count_letter_matches('RGBP', 'OROG') == 2
    assert count_letter_matches('RGBP', 'BWBR') == 2
    assert count_letter_matches('RBBP', 'GRBO') == 2
    assert count_letter_matches('RGBP', 'WOWO') == 0


def test_compare_codes():
    """
    Tests the student's compare_codes function; this function
    should always return a 4-character string composed of only
    'b', 'w', and '-' characters to represent a mastermind code.
    """
    # Check basic invariants.
    codes1 = make_random_codes(1000)
    codes2 = make_random_codes(1000)
    for (c1, c2) in zip(codes1, codes2):
        comp = compare_codes(c1, c2)
        # comparison should always be 4-character string
        assert len(comp) == 4
        for char in comp:
            # check that result only contains valid chars
            assert char in 'bw-'

    # Check specific answers.
    # exact match case
    assert compare_codes('RGBP', 'RGBP') == 'bbbb'
    # 4-character permutation case (same chars, all in different position)
    assert compare_codes('RGBP', 'GBPR') == 'wwww'
    assert compare_codes('RGBP', 'OROG') == 'ww--'
    # b should come before w before -
    assert compare_codes('RGBP', 'BWBR') == 'bw--'
    # all different
    assert compare_codes('RGBP', 'WOWO') == '----'

